from sklearn.cluster import AgglomerativeClustering
from .base import BaseModel

class AgglomerativeModel(BaseModel):
    """
    Agglomerative clustering model
    brute force search is being used
    """

    def __init(self, param_grid=None):
        super().__init__()
        self.param_grid = param_grid or {
            "n_clusters": [2, 3, 4, 5, 6],
            "linkage": ["ward", "complete", "average", "single"]
        }

    def train(self, X):
        best_score = -float("inf")
        best_model = None

        for n_clusters in self.param_grid["n_clusters"]:
            for linkage in self.param_grid["linkage"]:
                model = AgglomerativeClustering(
                    n_clusters=n_clusters,
                    linkage=linkage
                )
                labels = model.fit_predict(X)
                clusters = set(labels)
                score = 0
                for k in clusters:
                    cluster_size = (labels == k).sum()
                    score += 1 / cluster_size
                if score > best_score:
                    best_score = score
                    best_model = model
        
        self.best_model = best_model
        self.best_score = best_score
        return self.best_model
    


from sklearn.cluster import KMeans
from .base import BaseModel


class KMeansModel(BaseModel):
    """
    K-Means clustering model (unsupervised learning)
    Brute force method is being used
    """
    def __init__(self, param_grid=None):
        super().__init__()
        self.param_grid = param_grid or {
            "n_clusters": [2, 3, 4, 5, 6],
            "init": ["k-means++", "random"],
            "n_init": [10, 20],
            "max_iter": [300, 500]
        }

    def train(self, X):
        best_score = float("inf")
        best_model = None

        for n_clusters in self.param_grid["n_clusters"]:
            for init in self.param_grid["init"]:
                for n_init in self.param_grid["n_init"]:
                    for max_iter in self.param_grid["max_iter"]:
                        model = KMeans(
                            n_clusters=n_clusters,
                            init=init,
                            n_init=n_init,
                            max_iter=max_iter,
                            random_state=42
                        )
                        model.fit(X)
                        score = model.inertia_
                        if score < best_score:
                            best_score = score
                            best_model = model
        self.best_model = best_model
        self.best_score = best_score
        return self.best_model



from .base import BaseModel
from .utils import split_data
from .linear_regression import LinearRegressor
from .logistic_regression import LogisticClassifier
from .decision_tree import DecisionTreeModel
from .knn import KNNModel
from .random_forest import RandomForestModel
from .svm import SVMModel
from .naive_bayes import NaiveBayesModel
from .kmeans import KMeansModel
from .dbscan import DBSCANModel
from .agglomerative import AgglomerativeModel


from sklearn.cluster import DBSCAN
from .base import BaseModel

class DBSCANModel(BaseModel):
    """
    DBSCAN clustering model (unsupervised learning)
    brute force searching is being used
    """
    def __init__(self, param_grid=None):
        super().__init__()
        self.param_grid = param_grid or {
            "eps": [0.3, 0.5, 0.7, 1.0],
            "min_samples": [3, 5, 10]
        }

    def train(self, X):
        best_score = -float("inf")
        best_model = None

        for eps in self.param_grid["eps"]:
            for min_samples in self.param_grid["min_samples"]:
                model = DBSCAN(eps=eps, min_samples=min_samples)
                labels = model.fit_predict(X)

                score = len(labels) - (labels == -1).sum()
                if score > best_score:
                    best_score = score
                    best_model = model
        self.best_model = best_model
        self.best_score = best_score
        return self.best_model